/*
 *  Aquamarine the KDE window decorator
 *
 *  Copyright (c) 2006 Dennis Kasprzyk <onestone@beryl-project.org>
 *  Copyright (c) 2006 Volker Krause <vkrause@kde.org>
 *
 *  Uses code of:
 *      Emerald window decorator (www.beryl-project.org)
 *      gtk-window-decorator (www.freedesktop.org/wiki/Software/Compiz)
 *      KWin window manager (www.kde.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#ifndef AQUAMARINE_WINDOW_H
#define AQUAMARINE_WINDOW_H

#include <kdecoration_p.h>

#include <qpixmap.h>
#include <qwidget.h>

#include "utils.h"
#include "options.h"

class KDecoration;
class QPopupMenu;

namespace Aquamarine
{

    class Window:public QWidget, public KDecorationBridge
    {
      Q_OBJECT public:

        enum DecorationType
        {
            DecorationNormal,
            DecorationDefault,
            DecorationDefaultActive
        };

      public:
            Window (WId wid, WId frame, DecorationType type =
                    DecorationNormal);
           ~Window ();

        // KDecorationBridge methods
        virtual bool isActive () const;
        virtual bool isCloseable () const;
        virtual bool isMaximizable () const;
        virtual MaximizeMode maximizeMode () const;
        virtual bool isMinimizable () const;
        virtual bool providesContextHelp () const;
        virtual int desktop () const;
        virtual bool isModal () const;
        virtual bool isShadeable () const;
        virtual bool isShade () const;
        virtual bool isSetShade () const;
        virtual bool keepAbove () const;
        virtual bool keepBelow () const;
        virtual bool isMovable () const;
        virtual bool isResizable () const;
        virtual NET::
            WindowType windowType (unsigned long supported_types) const;
        virtual QIconSet icon () const;
        virtual QString caption () const;
        virtual void processMousePressEvent (QMouseEvent *);
        virtual void showWindowMenu (const QRect &);
        virtual void showWindowMenu (QPoint);
        virtual void performWindowOperation (WindowOperation);
        virtual void setMask (const QRegion &, int);
        virtual bool isPreview () const;
        virtual QRect geometry () const;
        virtual QRect iconGeometry () const;
        virtual QRegion unobscuredRegion (const QRegion & r) const;
        virtual QWidget *workspaceWidget () const;
        virtual WId windowId () const;
        virtual void closeWindow ();
        virtual void maximize (MaximizeMode mode);
        virtual void minimize ();
        virtual void showContextHelp ();
        virtual void setDesktop (int desktop);
        virtual void titlebarDblClickOperation ();
        virtual void titlebarMouseWheelOperation (int delta);
        virtual void setShade (bool set);
        virtual void setKeepAbove (bool);
        virtual void setKeepBelow (bool);

        virtual int currentDesktop () const;
        virtual QWidget *initialParentWidget () const;
        virtual Qt::WFlags initialWFlags () const;
        virtual void helperShowHide (bool);
        virtual void grabXServer (bool grab);


        void setActive (bool active);
        void updateFrame (WId frame);
        void updateWindowGeometry ();

        void updateCursor (QPoint pos);

        WId frameId () const
        {
            return m_frame;
        }
        KDecoration *decoration () const
        {
            return m_deco;
        }
        QWidget *activeChild () const
        {
            return m_activeChild;
        }
        void setActiveChild (QWidget * child)
        {
            m_activeChild = child;
        }

        void moveWindow (QMouseEvent * qme);

        void reloadDecoration ();

        void updateState ();
        void updateName ();
        void updateIcons ();

        void updateOpacity ()
        {
            m_opacity =
                readWindowPropertyDefault (m_wid, Atoms::opacity, 0xffff);
        }
        void updateBrightness ()
        {
            m_brightness =
                readWindowPropertyDefault (m_wid, Atoms::brightness, 0xffff);
        }
        void updateSaturation ()
        {
            m_saturation =
                readWindowPropertyDefault (m_wid, Atoms::saturation, 0xffff);
        }

      protected:
        virtual bool eventFilter (QObject * watched, QEvent * event);

      private:
        enum WindowAppearance
        {
            ResetValues,
            ChangeOpacity25,
            ChangeOpacity50,
            ChangeOpacity75,
            ChangeOpacity100,
            ChangeBrightness25,
            ChangeBrightness50,
            ChangeBrightness75,
            ChangeBrightness100,
            ChangeSaturation25,
            ChangeSaturation50,
            ChangeSaturation75,
            ChangeSaturation100
        };

      private:
        void createDecoration ();
        void resizeDecoration ( bool force = false );
        void getWindowProtocols ();
        void performMouseCommand( Options::MouseCommand, QMouseEvent *qme );

      private slots:
        void grabChildWidgets ();
        void updateWindowProperties ();
        void slotPopupActivated (int id);
        void slotPopupAboutToShow ();
        void slotChangeViewport (int id);
        void slotAppearanceChange (int id);
        void slotDesktopGeometryChanged ();

      private:
        WId m_wid;
        WId m_frame;

        // window properties
        QRect m_geometry;
        bool m_active;
        QString m_name;
        QIconSet m_icons;

        unsigned short m_opacity;
        unsigned short m_brightness;
        unsigned short m_saturation;

        // decoration stuff
        KDecoration *m_deco;
        QPixmap m_qpixmap;
        QPixmap m_pending_pixmap;
        bool m_fakeMaximized;

        // shape stuff
        bool m_shapeSet;
        QRegion m_shape;
        QRegion m_clip;

        // event handling stuff
        QWidget *m_activeChild;
        bool m_grabChildPending;
        bool m_grabChildInProgress;
        bool m_updateWindowPending;
        bool m_grabAllChildren;
        QValueList<QObject*> m_pendingChildren;

        // window handling
        bool m_supportDeleteWindow;
        bool m_supportTakeFocus;
        bool m_supportTakeActivity;
        bool m_supportContextHelp;
        bool m_supportPing;

        DecorationType m_type;

        // window menu
        QPopupMenu *m_popup;
        QPopupMenu *m_advanced_popup;
        QPopupMenu *m_viewport_menu;
        QPopupMenu *m_appearance_menu;
        QPopupMenu *m_brightness_menu;
        QPopupMenu *m_saturation_menu;
        QPopupMenu *m_opacity_menu;

        unsigned long m_state;
    };

}

#endif
